@@ -16,12 +16,13 @@
 
 #include "osmtypes.h"
 #include "output.h"
+#include "reprojection.h"
 #include "output-pgsql.h"
 #include "build_geometry.h"
 #include "middle-pgsql.h"
 
-/* Postgres database parameters */
-static char conninfo[256];
+/* Set if the output is in lat/lon */
+#define SRID (project_getprojinfo()->srs)
 
 enum table_id {
     t_point, t_line, t_poly, t_roads
@@ -33,10 +34,10 @@ static struct {
     const char *name;
     const char *type;
 } tables [] = {
-    { name: "planet_osm_point",   type: "POINT"     },
-    { name: "planet_osm_line",    type: "LINESTRING"},
-    { name: "planet_osm_polygon", type: "GEOMETRY"  },
-    { name: "planet_osm_roads",   type: "LINESTRING"}
+    { name: "%s_point",   type: "POINT"     },
+    { name: "%s_line",    type: "LINESTRING"},
+    { name: "%s_polygon", type: "POLYGON"  },
+    { name: "%s_roads",   type: "LINESTRING"}
 };
 static const unsigned int num_tables = sizeof(tables)/sizeof(*tables);
 
@@ -46,12 +47,16 @@ static struct {
     const char *type;
     const int polygon;
 } exportTags[] = {
+    {"access", "text", 0},
+    {"admin_level", "text", 0},
     {"aeroway",  "text", 1},
     {"amenity",  "text", 1},
     {"bicycle",  "text", 0},
     {"bridge",   "text", 0},
-    {"boundary", "text", 1},
+    {"boundary", "text", 0},
     {"building", "text", 1},
+    {"cutting",  "text", 0},
+    {"embankment","text", 0},
     {"foot",     "text", 0},
     {"highway",  "text", 0},
     {"horse",    "text", 0},
@@ -60,20 +65,25 @@ static struct {
     {"layer",    "text", 0},
     {"learning", "text", 0},
     {"leisure",  "text", 1},
-    {"man_made", "text", 0},
+    {"man_made", "text", 1},
+    {"military", "text", 1},
     {"motorcar", "text", 0},
     {"name",     "text", 0},
     {"natural",  "text", 1},
     {"oneway",   "text", 0},
+    {"power",    "text", 1},
     {"place",    "text", 0},
     {"railway",  "text", 0},
     {"ref",      "text", 0},
+    {"religion", "text", 0},
     {"residence","text", 0},
     {"route",    "text", 0},
     {"sport",    "text", 1},
     {"tourism",  "text", 1},
     {"tunnel",   "text", 0},
     {"waterway", "text", 0},
+    {"width",    "text", 0},
+    {"wood",     "text", 0},
     {"z_order",  "int4", 0},
     {"way_area", "real", 0}
 };
@@ -154,6 +164,7 @@ static int add_z_order_line(struct keyval *tags, int *roads)
     const char *layer   = getItem(tags, "layer");
     const char *highway = getItem(tags, "highway");
     const char *bridge  = getItem(tags, "bridge");
+    const char *tunnel  = getItem(tags, "tunnel");
     const char *railway = getItem(tags, "railway");
     int z_order = 0;
     int l;
@@ -182,6 +193,9 @@ static int add_z_order_line(struct keyval *tags, int *roads)
     if (bridge && (!strcmp(bridge, "true") || !strcmp(bridge, "yes") || !strcmp(bridge, "1")))
         z_order += 10;
 
+    if (tunnel && (!strcmp(tunnel, "true") || !strcmp(tunnel, "yes") || !strcmp(tunnel, "1")))
+        z_order -= 10;
+
     snprintf(z, sizeof(z), "%d", z_order);
     addItem(tags, "z_order", z, 0);
 
@@ -313,7 +327,7 @@ static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double n
         sql_out(sql_conn, "\t", 1);
     }
 
-    sprintf(sql, "SRID=4326;POINT(%.15g %.15g)", node_lon, node_lat);
+    sprintf(sql, "SRID=%d;POINT(%.15g %.15g)", SRID, node_lon, node_lat);
     sql_out(sql_conn, sql, strlen(sql));
     sql_out(sql_conn, "\n", 1);
 
@@ -341,35 +355,58 @@ static void write_wkts(int id, struct keyval *tags, const char *wkt, PGconn *sql
 	    sql_out(sql_conn, "\t", 1);
     }
 
-    sprintf(sql, "SRID=4326;");
+    sprintf(sql, "SRID=%d;", SRID);
     sql_out(sql_conn, sql, strlen(sql));
     sql_out(sql_conn, wkt, strlen(wkt));
     sql_out(sql_conn, "\n", 1);
 }
 
-void add_parking_node(int id, struct keyval *tags)
+void add_parking_node(int id, struct keyval *tags, double node_lat, double node_lon)
 {
 // insert into planet_osm_point(osm_id,name,amenity,way) select osm_id,name,amenity,centroid(way) from planet_osm_polygon where amenity='parking';
+	const char *access  = getItem(tags, "access");
 	const char *amenity = getItem(tags, "amenity");
 	const char *name    = getItem(tags, "name");
 	struct keyval nodeTags;
-	double node_lat = 0, node_lon = 0;
 	
 	if (!amenity || strcmp(amenity, "parking"))
 		return;
 
+	// Do not add a 'P' symbol if access is defined and something other than public.
+	if (access && strcmp(access, "public"))
+		return;
+
 	initList(&nodeTags);
 	addItem(&nodeTags, "amenity", amenity, 0);
 	if (name)
 		addItem(&nodeTags, "name",    name,    0);
-	
-	//fprintf(stderr, "Parking node: %s\t%f,%f\n", name ? name : "no_name", node_lat, node_lon);
+	if (access)
+		addItem(&nodeTags, "access",  access,    0);
 	
 	pgsql_out_node(id, &nodeTags, node_lat, node_lon);
 	resetList(&nodeTags);
 }
 
 
+unsigned int pgsql_filter_tags(struct keyval *tags, int *polygon)
+{
+    const char *v;
+    unsigned int i, filter = 1;
+
+    *polygon = 0;
+
+    for (i=0; i < numTags; i++) {
+        if ((v = getItem(tags, exportTags[i].name))) {
+            filter = 0;
+            *polygon |= exportTags[i].polygon;
+            if (*polygon)
+                break;
+        }
+    }
+
+    return filter;
+}
+
 /*
 COPY planet_osm (osm_id, name, place, landuse, leisure, "natural", man_made, waterway, highway, railway, amenity, tourism, learning, bu
 ilding, bridge, layer, way) FROM stdin;
@@ -380,33 +417,23 @@ E4C1421D5BF24D06053E7DF4940
 */
 static int pgsql_out_way(int id, struct keyval *tags, struct osmNode *nodes, int count)
 {
-    const char *v;
-    unsigned int i;
-    int polygon = 0, export = 0;
-    int roads = 0;
+    int polygon = 0, roads = 0;
     char *wkt;
-    double area;
+    double area, interior_lat, interior_lon;
+    const char *multipolygon = getItem(tags, "multipolygon");
 
-    for (i=0; i < numTags; i++) {
-        if ((v = getItem(tags, exportTags[i].name))) {
-            export = 1;
-            polygon |= exportTags[i].polygon;
-            if (polygon)
-                break;
-        }
-    }
-
-    if (!export)
+    // If the way has been part of a multipolygon then skip
+    if (multipolygon)
         return 0;
 
-    if (add_z_order(tags, polygon, &roads))
+    if (pgsql_filter_tags(tags, &polygon) || add_z_order(tags, polygon, &roads))
         return 0;
 
     //compress_tag_name(tags);
 
     fix_motorway_shields(tags);
 
-    wkt = get_wkt(nodes, count, polygon, &area);
+    wkt = get_wkt_simple(nodes, count, polygon, &area, &interior_lon, &interior_lat);
     if (wkt && strlen(wkt)) {
 	/* FIXME: there should be a better way to detect polygons */
 	if (!strncmp(wkt, "POLYGON", strlen("POLYGON"))) {
@@ -416,7 +443,7 @@ static int pgsql_out_way(int id, struct keyval *tags, struct osmNode *nodes, int
 		addItem(tags, "way_area", tmp, 0);
 	    }
 	    write_wkts(id, tags, wkt, sql_conns[t_poly]);
-	    add_parking_node(id, tags);
+	    add_parking_node(id, tags, interior_lat, interior_lon);
 	} else {
 	    write_wkts(id, tags, wkt, sql_conns[t_line]);
 	    if (roads)
@@ -428,15 +455,87 @@ static int pgsql_out_way(int id, struct keyval *tags, struct osmNode *nodes, int
     return 0;
 }
 
+static int pgsql_out_relation(int id, struct keyval *rel_tags, struct osmNode **xnodes, struct keyval **xtags, int *xcount)
+{
+    unsigned int i, wkt_size;
+    double interior_lat, interior_lon;
+    int polygon = 0, roads = 0;
+    struct keyval tags, *p;
+#if 0
+    fprintf(stderr, "Got relation with counts:");
+    for (i=0; xcount[i]; i++)
+        fprintf(stderr, " %d", xcount[i]);
+    fprintf(stderr, "\n");
+#endif
+    // Aggregate tags from all polygons with the relation,
+    // no idea is this is a good way to deal with differing tags across the ways or not
+    initList(&tags);
+    p = rel_tags->next;
+    while (p != rel_tags) {
+        addItem(&tags, p->key, p->value, 1);
+        p = p->next;
+    }
+    for (i=0; xtags[i]; i++) {
+        p = xtags[i]->next;
+        while (p != xtags[i]) {
+            addItem(&tags, p->key, p->value, 1);
+            p = p->next;
+        }
+    }
 
-static int pgsql_out_start(const char *db, int append)
+    if (pgsql_filter_tags(&tags, &polygon) || add_z_order(&tags, polygon, &roads)) {
+        resetList(&tags);
+        return 0;
+    }
+
+    wkt_size = build_geometry(id, xnodes, xcount);
+
+    if (!wkt_size) {
+        resetList(&tags);
+        return 0;
+    }
+
+    for (i=0;i<wkt_size;i++)
+    {
+        char *wkt = get_wkt(i);
+
+        if (strlen(wkt)) {
+            /* FIXME: there should be a better way to detect polygons */
+            if (!strncmp(wkt, "POLYGON", strlen("POLYGON"))) {
+                double area = get_area(i);
+                if (area > 0.0) {
+                    char tmp[32];
+                    snprintf(tmp, sizeof(tmp), "%f", area);
+                    addItem(&tags, "way_area", tmp, 0);
+                }
+                write_wkts(id, &tags, wkt, sql_conns[t_poly]);
+                get_interior(i, &interior_lat, &interior_lon);
+                add_parking_node(id, &tags, interior_lat, interior_lon);
+	    } else {
+                write_wkts(id, &tags, wkt, sql_conns[t_line]);
+                if (roads)
+                    write_wkts(id, &tags, wkt, sql_conns[t_roads]);
+            }
+	}
+        free(wkt);
+    }
+	
+    clear_wkts();
+
+    // Mark each member so that we can skip them during iterate_ways
+    for (i=0; xcount[i]; i++)
+        addItem(xtags[i], "multipolygon", "1", 0);
+
+    resetList(&tags);
+    return 0;
+}
+
+static int pgsql_out_start(const char *conninfo, const char *prefix, int append)
 {
     char sql[1024], tmp[128];
     PGresult   *res;
     unsigned int i,j;
 
-    snprintf(conninfo, sizeof(conninfo), "dbname = %s", db);
-
     /* We use a connection per table to enable the use of COPY_IN */
     sql_conns = calloc(num_tables, sizeof(PGconn *));
     assert(sql_conns);
@@ -444,6 +543,12 @@ static int pgsql_out_start(const char *db, int append)
     for (i=0; i<num_tables; i++) {
         PGconn *sql_conn;
 
+        /* Substitute prefix into name of table */
+        {
+            char *temp = malloc( strlen(prefix) + strlen(tables[i].name) + 1 );
+            sprintf( temp, tables[i].name, prefix );
+            tables[i].name = temp;
+        }
         fprintf(stderr, "Setting up table: %s\n", tables[i].name);
         sql_conn = PQconnectdb(conninfo);
 
@@ -455,9 +560,7 @@ static int pgsql_out_start(const char *db, int append)
         sql_conns[i] = sql_conn;
 
         if (!append) {
-            sql[0] = '\0';
-            strcat(sql, "DROP TABLE ");
-            strcat(sql, tables[i].name);
+            sprintf( sql, "DROP TABLE %s;", tables[i].name);
             res = PQexec(sql_conn, sql);
             PQclear(res); /* Will be an error if table does not exist */
         }
@@ -471,23 +574,19 @@ static int pgsql_out_start(const char *db, int append)
         PQclear(res);
 
         if (!append) {
-            sql[0] = '\0';
-            strcat(sql, "CREATE TABLE ");
-            strcat(sql, tables[i].name);
-            strcat(sql, " ( osm_id int4");
+            sprintf(sql, "CREATE TABLE %s ( osm_id int4", tables[i].name );
             for (j=0; j < numTags; j++) {
                 sprintf(tmp, ",\"%s\" %s", exportTags[j].name, exportTags[j].type);
                 strcat(sql, tmp);
             }
             strcat(sql, " );\n");
-            strcat(sql, "SELECT AddGeometryColumn('");
-            strcat(sql, tables[i].name);
-            strcat(sql, "', 'way', 4326, '");
-            strcat(sql, tables[i].type);
-            strcat(sql, "', 2 );\n");
+            sprintf( sql + strlen(sql), "SELECT AddGeometryColumn('%s', 'way', %d, '%s', 2 );\n",
+                        tables[i].name, SRID, tables[i].type );
+
+            sprintf( sql + strlen(sql), "ALTER TABLE %s ALTER COLUMN way SET NOT NULL;\n", tables[i].name);
 
             res = PQexec(sql_conn, sql);
-            if (PQresultStatus(res) != PGRES_TUPLES_OK) {
+            if (PQresultStatus(res) != PGRES_COMMAND_OK) {
                 fprintf(stderr, "%s failed: %s\n", sql, PQerrorMessage(sql_conn));
                 PQclear(res);
                 exit_nicely();
@@ -496,9 +595,7 @@ static int pgsql_out_start(const char *db, int append)
         }
 
         sql[0] = '\0';
-        strcat(sql, "COPY ");
-        strcat(sql, tables[i].name);
-        strcat(sql, " FROM STDIN");
+        sprintf(sql, "COPY %s FROM STDIN", tables[i].name);
 
         res = PQexec(sql_conn, sql);
         if (PQresultStatus(res) != PGRES_COPY_IN) {
@@ -512,7 +609,8 @@ static int pgsql_out_start(const char *db, int append)
     return 0;
 }
 
-static void pgsql_out_stop(int append)
+#define __unused  __attribute__ ((unused))
+static void pgsql_out_stop(__unused int append)
 {
     char sql[1024], tmp[128];
     PGresult   *res;
@@ -548,38 +646,33 @@ static void pgsql_out_stop(int append)
         PQclear(res);
 
         sql[0] = '\0';
-        strcat(sql, "VACUUM ANALYZE ");
+        strcat(sql, "ANALYZE ");
         strcat(sql, tables[i].name);
         strcat(sql, ";\n");
 
-        if (!append) {
-            strcat(sql, "ALTER TABLE ");
-            strcat(sql, tables[i].name);
-            strcat(sql, " ALTER COLUMN way SET NOT NULL;\n");
+        strcat(sql, "CREATE TABLE tmp AS SELECT * FROM ");
+        strcat(sql, tables[i].name);
+        strcat(sql, " ORDER BY way;\n");
 
-            strcat(sql, "CREATE INDEX way_index");
-            strcat(sql, tmp);
-            strcat(sql, " ON ");
-            strcat(sql, tables[i].name);
-            strcat(sql, " USING GIST (way GIST_GEOMETRY_OPS);\n");
-
-            strcat(sql, "CREATE INDEX z_index");
-            strcat(sql, tmp);
-            strcat(sql, " ON ");
-            strcat(sql, tables[i].name);
-            strcat(sql, " (z_order);\n");
-        }
-        strcat(sql, "CLUSTER way_index");
-        strcat(sql, tmp);
-        strcat(sql, " ON ");
+        strcat(sql, "DROP TABLE ");
         strcat(sql, tables[i].name);
         strcat(sql, ";\n");
 
+        strcat(sql, "ALTER TABLE tmp RENAME TO ");
+        strcat(sql, tables[i].name);
+        strcat(sql, ";\n");
+
+        strcat(sql, "CREATE INDEX ");
+        strcat(sql, tables[i].name);
+        strcat(sql, "_index ON ");
+        strcat(sql, tables[i].name);
+        strcat(sql, " USING GIST (way GIST_GEOMETRY_OPS);\n");
+
         strcat(sql, "GRANT SELECT ON ");
         strcat(sql, tables[i].name);
         strcat(sql, " TO PUBLIC;\n");
 
-        strcat(sql, "VACUUM ANALYZE ");
+        strcat(sql, "ANALYZE ");
         strcat(sql, tables[i].name);
         strcat(sql, ";\n");
 
@@ -601,5 +694,6 @@ struct output_t out_pgsql = {
         stop:      pgsql_out_stop,
         cleanup:   pgsql_out_cleanup,
         node:      pgsql_out_node,
-        way:       pgsql_out_way
+        way:       pgsql_out_way,
+        relation:  pgsql_out_relation
 };
